"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ViewportList = void 0;
var jsx_runtime_1 = require("react/jsx-runtime");
var react_1 = require("react");
var MIN_INDEX = 0;
var IS_TOUCH = (function () {
    try {
        return 'ontouchstart' in window || navigator.maxTouchPoints;
    }
    catch (error) {
        return false;
    }
})();
var IS_OVERFLOW_ANCHOR_SUPPORTED = (function () {
    try {
        return window.CSS.supports('overflow-anchor: auto');
    }
    catch (error) {
        return false;
    }
})();
var PROP_NAME_FOR_Y_AXIS = {
    top: 'top',
    bottom: 'bottom',
    clientHeight: 'clientHeight',
    scrollHeight: 'scrollHeight',
    scrollTop: 'scrollTop',
    overflowY: 'overflowY',
    height: 'height',
    minHeight: 'minHeight',
    maxHeight: 'maxHeight',
    marginTop: 'marginTop'
};
var PROP_NAME_FOR_X_AXIS = {
    top: 'left',
    bottom: 'right',
    clientHeight: 'clientWidth',
    scrollHeight: 'scrollWidth',
    scrollTop: 'scrollLeft',
    overflowY: 'overflowX',
    height: 'width',
    minHeight: 'minWidth',
    maxHeight: 'maxWidth',
    marginTop: 'marginLeft'
};
var getStyle = function (propName, size, marginTop) {
    var _a;
    if (marginTop === void 0) { marginTop = 0; }
    return (_a = {},
        _a[propName.minHeight] = size,
        _a[propName.height] = size,
        _a[propName.maxHeight] = size,
        _a.overflowAnchor = 'none',
        _a.pointerEvents = 'none',
        _a.userSelect = 'none',
        _a.padding = 0,
        _a.margin = 0,
        _a.border = 'none',
        _a[propName.marginTop] = marginTop,
        _a);
};
var normalizeValue = function (min, value, max) { return Math.max(Math.min(value, max), min); };
var useIsomorphicLayoutEffect = typeof window !== 'undefined' ? react_1.useLayoutEffect : react_1.useEffect;
var useRequestAnimationFrame = function (func) {
    var stepRef = (0, react_1.useRef)(func);
    stepRef.current = func;
    (0, react_1.useEffect)(function () {
        var frameId;
        var frame = function () {
            frameId = requestAnimationFrame(frame);
            stepRef.current();
        };
        frame();
        return function () {
            cancelAnimationFrame(frameId);
        };
    }, []);
};
var ViewportListInner = function (_a, ref) {
    var viewportRef = _a.viewportRef, _b = _a.items, items = _b === void 0 ? [] : _b, _c = _a.itemMinSize, itemMinSize = _c === void 0 ? 1 : _c, _d = _a.margin, margin = _d === void 0 ? 0 : _d, _e = _a.overscan, overscan = _e === void 0 ? 1 : _e, _f = _a.axis, axis = _f === void 0 ? 'y' : _f, _g = _a.initialIndex, initialIndex = _g === void 0 ? 0 : _g, _h = _a.initialAlignToTop, initialAlignToTop = _h === void 0 ? true : _h, _j = _a.initialOffset, initialOffset = _j === void 0 ? 0 : _j, _k = _a.fixed, fixed = _k === void 0 ? false : _k, children = _a.children;
    var propName = axis === 'y' ? PROP_NAME_FOR_Y_AXIS : PROP_NAME_FOR_X_AXIS;
    var maxIndex = items.length - 1;
    var normalizedItemMinSize = Math.max(1, itemMinSize);
    var itemMinSizeWithMargin = normalizedItemMinSize + margin;
    var overscanSize = overscan * itemMinSizeWithMargin;
    var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
    var _l = (0, react_1.useState)([normalizedInitialIndex, normalizedInitialIndex]), _m = _l[0], startIndex = _m[0], endIndex = _m[1], setIndexes = _l[1];
    var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
    var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
    var topSpacerRef = (0, react_1.useRef)(null);
    var bottomSpacerRef = (0, react_1.useRef)(null);
    var cacheRef = (0, react_1.useRef)([]);
    var scrollToIndexRef = (0, react_1.useRef)(normalizedStartIndex
        ? { index: normalizedStartIndex, alignToTop: initialAlignToTop, offset: initialOffset }
        : null);
    var anchorIndexRef = (0, react_1.useRef)(-1);
    var marginTopRef = (0, react_1.useRef)(0);
    var topSpacerStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(MIN_INDEX, normalizedStartIndex)
            .reduce(function (sum, next) { return sum + (next - normalizedItemMinSize); }, normalizedStartIndex * itemMinSizeWithMargin), marginTopRef.current);
    }, [normalizedItemMinSize, itemMinSizeWithMargin, normalizedStartIndex, propName]);
    var bottomSpacerStyle = (0, react_1.useMemo)(function () {
        return getStyle(propName, cacheRef.current
            .slice(normalizedEndIndex + 1, maxIndex + 1)
            .reduce(function (sum, next) { return sum + (next - normalizedItemMinSize); }, itemMinSizeWithMargin * (maxIndex - normalizedEndIndex)));
    }, [propName, normalizedEndIndex, maxIndex, itemMinSizeWithMargin, normalizedItemMinSize]);
    useRequestAnimationFrame(function () {
        var viewport = viewportRef ? viewportRef.current : document.documentElement;
        var topSpacer = topSpacerRef.current;
        var bottomSpacer = bottomSpacerRef.current;
        if (!viewport || !topSpacer || !bottomSpacer) {
            return;
        }
        if (marginTopRef.current && (viewport.scrollTop <= 0 || normalizedStartIndex === 0)) {
            topSpacer.style[propName.marginTop] = '0px';
            viewport.style[propName.overflowY] = 'hidden';
            viewport.scrollTop += -marginTopRef.current;
            viewport.style[propName.overflowY] = '';
            marginTopRef.current = 0;
            return;
        }
        var topElement = topSpacer.nextSibling;
        var bottomElement = bottomSpacer.previousSibling;
        if (!topElement || !bottomElement) {
            return;
        }
        var viewportRect = viewport.getBoundingClientRect();
        var topElementRect = topElement.getBoundingClientRect();
        var bottomElementRect = bottomElement.getBoundingClientRect();
        var topLimit = normalizeValue(0, viewportRect[propName.top], document.documentElement[propName.clientHeight]) -
            overscanSize;
        var bottomLimit = normalizeValue(0, viewportRect[propName.bottom], document.documentElement[propName.clientHeight]) +
            overscanSize;
        var maxItemsCountInViewPort = normalizeValue(0, Math.ceil((bottomLimit - topLimit) / itemMinSizeWithMargin), items.length);
        var nextStartIndex = normalizedStartIndex;
        var nextEndIndex = normalizedEndIndex;
        if (scrollToIndexRef.current) {
            var targetIndex = normalizeValue(MIN_INDEX, scrollToIndexRef.current.index, maxIndex);
            if (targetIndex >= normalizedStartIndex && targetIndex <= normalizedEndIndex) {
                var index = normalizedStartIndex;
                var element = topElement;
                while (element && element !== bottomSpacer) {
                    if (index === targetIndex) {
                        element.scrollIntoView(scrollToIndexRef.current.alignToTop);
                        if (scrollToIndexRef.current.offset) {
                            viewport[propName.scrollTop] += scrollToIndexRef.current.offset;
                        }
                        scrollToIndexRef.current = null;
                        return;
                    }
                    index++;
                    element = element.nextSibling;
                }
                return;
            }
            nextStartIndex = targetIndex - maxItemsCountInViewPort;
            nextEndIndex = targetIndex + maxItemsCountInViewPort;
        }
        else if (fixed) {
            nextStartIndex = Math.floor((topLimit - topSpacerRef.current.getBoundingClientRect()[propName.top]) / itemMinSizeWithMargin);
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else if (topElementRect[propName.top] >= bottomLimit) {
            // fast scroll up
            var diff = topElementRect[propName.top] - bottomLimit;
            nextEndIndex = normalizedStartIndex;
            while (diff >= 0 && nextEndIndex > MIN_INDEX) {
                nextEndIndex--;
                diff -= (cacheRef.current[nextEndIndex] || normalizedItemMinSize) + margin;
            }
            anchorIndexRef.current = normalizedStartIndex;
            nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
        }
        else if (bottomElementRect[propName.bottom] + margin <= topLimit) {
            // fast scroll down
            var diff = topLimit - bottomElementRect[propName.bottom] + margin;
            nextStartIndex = normalizedEndIndex;
            while (diff >= 0 && nextStartIndex < maxIndex) {
                nextStartIndex++;
                diff -= (cacheRef.current[nextStartIndex] || normalizedItemMinSize) + margin;
            }
            nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
        }
        else {
            if (topElementRect[propName.bottom] + margin < topLimit) {
                // scroll down (correction)
                nextStartIndex++;
            }
            else if (topElementRect[propName.top] >= topLimit) {
                // scroll up
                var diff = topElementRect[propName.top] - topLimit;
                while (diff >= 0 && nextStartIndex > MIN_INDEX) {
                    nextStartIndex--;
                    diff -= (cacheRef.current[nextStartIndex] || normalizedItemMinSize) + margin;
                }
                anchorIndexRef.current = normalizedStartIndex;
            }
            if (bottomElementRect[propName.bottom] + margin <= bottomLimit) {
                // scroll down
                var diff = bottomLimit - bottomElementRect[propName.bottom] - margin;
                while (diff >= 0 && nextEndIndex < maxIndex) {
                    nextEndIndex++;
                    diff -= (cacheRef.current[nextEndIndex] || normalizedItemMinSize) + margin;
                }
            }
            else if (bottomElementRect[propName.top] > bottomLimit) {
                // scroll up (correction)
                nextEndIndex--;
            }
        }
        nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
        nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);
        if (nextStartIndex !== normalizedStartIndex || nextEndIndex !== normalizedEndIndex) {
            if (fixed) {
                cacheRef.current = [];
            }
            else {
                var index = normalizedStartIndex;
                var element = topElement;
                while (element && index < nextStartIndex && element !== bottomSpacer) {
                    cacheRef.current[index] = element[propName.clientHeight];
                    index++;
                    element = element.nextSibling;
                }
                index = normalizedEndIndex;
                element = bottomElement;
                while (element && index > nextEndIndex && element !== topSpacer) {
                    cacheRef.current[index] = element[propName.clientHeight];
                    index--;
                    element = element.previousSibling;
                }
            }
            setIndexes([nextStartIndex, nextEndIndex]);
        }
    });
    useIsomorphicLayoutEffect(function () {
        var anchorIndex = anchorIndexRef.current;
        anchorIndexRef.current = -1;
        var viewport = viewportRef ? viewportRef.current : document.documentElement;
        var topSpacer = topSpacerRef.current;
        var bottomSpacer = bottomSpacerRef.current;
        if (IS_OVERFLOW_ANCHOR_SUPPORTED || anchorIndex === -1 || !viewport || !topSpacer || !bottomSpacer) {
            return;
        }
        var topElement = topSpacer.nextSibling;
        if (!topElement) {
            return;
        }
        var index = normalizedStartIndex;
        var element = topElement;
        var sizeDiff = 0;
        while (element && index < anchorIndex && element !== bottomSpacer) {
            sizeDiff += element[propName.clientHeight] - (cacheRef.current[index] || normalizedItemMinSize);
            index++;
            element = element.nextSibling;
        }
        if (sizeDiff) {
            if (IS_TOUCH) {
                marginTopRef.current -= sizeDiff;
                topSpacer.style[propName.marginTop] = marginTopRef.current + "px";
            }
            else {
                viewport[propName.scrollTop] += sizeDiff;
            }
        }
        anchorIndexRef.current = -1;
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, [normalizedStartIndex]);
    (0, react_1.useImperativeHandle)(ref, function () { return ({
        scrollToIndex: function (index, alignToTop, offset) {
            if (index === void 0) { index = -1; }
            if (alignToTop === void 0) { alignToTop = true; }
            if (offset === void 0) { offset = 0; }
            scrollToIndexRef.current = { index: index, alignToTop: alignToTop, offset: offset };
        }
    }); }, []);
    return ((0, jsx_runtime_1.jsxs)(react_1.Fragment, { children: [(0, jsx_runtime_1.jsx)("div", { ref: topSpacerRef, style: topSpacerStyle }, void 0), items
                .slice(normalizedStartIndex, normalizedEndIndex + 1)
                .map(function (item, index) { return children(item, normalizedStartIndex + index, items); }), (0, jsx_runtime_1.jsx)("div", { ref: bottomSpacerRef, style: bottomSpacerStyle }, void 0)] }, void 0));
};
exports.ViewportList = (0, react_1.forwardRef)(ViewportListInner);
// eslint-disable-next-line import/no-default-export
exports.default = exports.ViewportList;
